import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    KARABO_CLASSINFO,
    NDARRAY_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Schema,
)
from karabo.common.states import State

from . import shmem_utils, utils
from ._version import version as deviceVersion
from .agipd_gpu import AgipdGainMode, AgipdGpuRunner, BadPixelValues, CorrectionFlags
from .base_correction import BaseCorrection, add_correction_step_schema
from .calcat_utils import AgipdCalcatFriend, AgipdConstants


@KARABO_CLASSINFO("AgipdCorrection", deviceVersion)
class AgipdCorrection(BaseCorrection):
    # subclass *must* set these attributes
    _correction_flag_class = CorrectionFlags
    _correction_field_names = (
        ("thresholding", CorrectionFlags.THRESHOLD),
        ("offset", CorrectionFlags.OFFSET),
        ("relGainPc", CorrectionFlags.REL_GAIN_PC),
        ("relGainXray", CorrectionFlags.REL_GAIN_XRAY),
        ("badPixels", CorrectionFlags.BPMASK),
    )
    _gpu_runner_class = AgipdGpuRunner
    _calcat_friend_class = AgipdCalcatFriend
    _constant_enum_class = AgipdConstants
    _managed_keys = BaseCorrection._managed_keys | {
        "sendGainMap",
    }

    # this is just extending (not mandatory)
    _schema_cache_fields = BaseCorrection._schema_cache_fields | {
        "sendGainMap",
    }

    @staticmethod
    def expectedParameters(expected):
        super(AgipdCorrection, AgipdCorrection).expectedParameters(expected)
        expected.setDefaultValue("dataFormat.memoryCells", 352)
        expected.setDefaultValue("preview.selectionMode", "cell")
        (
            STRING_ELEMENT(expected)
            .key("gainMode")
            .displayedName("Gain mode")
            .assignmentOptional()
            .defaultValue("ADAPTIVE_GAIN")
            .options("ADAPTIVE_GAIN,FIXED_HIGH_GAIN,FIXED_MEDIUM_GAIN,FIXED_LOW_GAIN")
            .commit(),
            BOOL_ELEMENT(expected)
            .key("sendGainMap")
            .displayedName("Send gain map on dataOutput")
            .assignmentOptional()
            .defaultValue(False)
            .commit(),
        )
        # TODO: make sendGainMap reconfigurable

        (
            STRING_ELEMENT(expected)
            .key("dataOutput.schema.image.gainMap")
            .displayedName("Gain map (optional)")
            .assignmentOptional()
            .defaultValue("")
            .commit()
        )
        preview_schema = Schema()
        (
            NODE_ELEMENT(preview_schema).key("data").commit(),
            NDARRAY_ELEMENT(preview_schema).key("data.adc").dtype("FLOAT").commit(),
        )
        (
            OUTPUT_CHANNEL(expected)
            .key("preview.outputGain")
            .dataSchema(preview_schema)
            .commit(),
        )

        AgipdCalcatFriend.add_schema(expected, AgipdCorrection._managed_keys)
        # this is not automatically done by superclass for complicated class reasons
        add_correction_step_schema(
            expected,
            AgipdCorrection._managed_keys,
            AgipdCorrection._correction_field_names,
        )

        # additional settings specific to AGIPD correction steps
        (
            BOOL_ELEMENT(expected)
            .key("corrections.relGainPc.overrideMdAdditionalOffset")
            .displayedName("Override md_additional_offset")
            .description(
                "Toggling this on will use the value in the next field globally for "
                "md_additional_offset. Note that the correction map on GPU gets "
                "overwritten as long as this boolean is True, so reload constants "
                "after turning off."
            )
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("corrections.relGainPc.mdAdditionalOffset")
            .displayedName("Value for md_additional_offset (if overriding)")
            .description(
                "Normally, md_additional_offset (part of relative gain correction) is "
                "computed when loading SlopesPC. In case you want to use a different "
                "value (global for all medium gain pixels), you can specify it here "
                "and set corrections.overrideMdAdditionalOffset to True."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("corrections.relGainXray.gGainValue")
            .displayedName("G_gain_value")
            .description(
                "Newer X-ray gain correction constants are absolute. The default "
                "G_gain_value of 1 means that output is expected to be in keV. If "
                "this is not desired, one can here specify the mean X-ray gain value "
                "over all modules to get ADU values out - operator must manually "
                "find this mean value."
            )
            .assignmentOptional()
            .defaultValue(1)
            .reconfigurable()
            .commit(),
            STRING_ELEMENT(expected)
            .key("corrections.badPixels.maskingValue")
            .displayedName("Bad pixel masking value")
            .description(
                "Any pixels masked by the bad pixel mask will have their value "
                "replaced with this. Note that this setting is evaluated as a string; "
                "use float('nan') to get NaN value."
            )
            .assignmentOptional()
            .defaultValue("float('nan')")
            .reconfigurable()
            .commit(),
            NODE_ELEMENT(expected)
            .key("corrections.badPixels.subsetToUse")
            .displayedName("Bad pixel flags to use")
            .description(
                "The booleans under this node allow for selecting a subset of bad "
                "pixel types to take into account when doing bad pixel masking. "
                "Upon updating these flags, the map used for bad pixel masking will "
                "be ANDed with this selection. Turning disabled flags back on causes "
                "reloading of cached constants."
            )
            .commit(),
        )
        AgipdCorrection._managed_keys.add(
            "corrections.relGainPc.overrideMdAdditionalOffset"
        )
        AgipdCorrection._managed_keys.add("corrections.relGainPc.mdAdditionalOffset")
        AgipdCorrection._managed_keys.add("corrections.relGainXray.gGainValue")
        AgipdCorrection._managed_keys.add("corrections.badPixels.maskingValue")
        # TODO: DRY / encapsulate
        for field in BadPixelValues:
            (
                BOOL_ELEMENT(expected)
                .key(f"corrections.badPixels.subsetToUse.{field.name}")
                .assignmentOptional()
                .defaultValue(True)
                .reconfigurable()
                .commit()
            )
            AgipdCorrection._managed_keys.add(
                f"corrections.badPixels.subsetToUse.{field.name}"
            )

        # mandatory: manager needs this in schema
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(AgipdCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self._schema_cache["dataFormat.memoryCells"],
            2,
            self._schema_cache["dataFormat.pixelsX"],
            self._schema_cache["dataFormat.pixelsY"],
        )

    def __init__(self, config):
        super().__init__(config)
        # TODO: different gpu runner for fixed gain mode
        self.gain_mode = AgipdGainMode[config.get("gainMode")]
        self.bad_pixel_mask_value = eval(
            config.get("corrections.badPixels.maskingValue")
        )
        self._gpu_runner_init_args = {
            "gain_mode": self.gain_mode,
            "bad_pixel_mask_value": self.bad_pixel_mask_value,
            "g_gain_value": config.get("corrections.relGainXray.gGainValue"),
            "output_gain_map": config.get("sendGainMap"),
        }

        self._shmem_buffer_gain_map = None

        # configurability: overriding md_additional_offset
        if config.get("corrections.relGainPc.overrideMdAdditionalOffset"):
            self._override_md_additional_offset = config.get(
                "corrections.relGainPc.mdAdditionalOffset"
            )
        else:
            self._override_md_additional_offset = None

        # configurability: disabling subset of bad pixel masking bits
        self._has_updated_bad_pixel_selection = False
        self.registerInitialFunction(self._update_bad_pixel_selection)

        self.updateState(State.ON)

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        do_generate_preview,
    ):
        """Called by input_handler for each data hash. Should correct data, optionally
        compute preview, write data output, and optionally write preview outputs."""
        # original shape: memory_cell, data/raw_gain, x, y

        # TODO: add pulse filter back in
        pulse_table = np.squeeze(data_hash.get("image.pulseId"))

        try:
            self.gpu_runner.load_data(image_data)
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.gpu_runner.load_cell_table(cell_table)
        self.gpu_runner.correct(self._correction_flag_enabled)
        self.gpu_runner.reshape(
            output_order=self._schema_cache["dataFormat.outputAxisOrder"],
            out=buffer_array,
        )
        # after reshape, data for dataOutput is now safe in its own buffer
        if do_generate_preview:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.gpu_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ) = utils.pick_frame_index(
                self._schema_cache["preview.selectionMode"],
                self._schema_cache["preview.index"],
                cell_table,
                pulse_table,
                warn_func=self.log_status_warn,
            )
            (
                preview_raw,
                preview_corrected,
            ) = self.gpu_runner.compute_preview(preview_slice_index)
            if self._schema_cache["sendGainMap"]:
                preview_gain = self.gpu_runner.compute_preview_gain(
                    preview_slice_index
                )

        # reusing input data hash for sending
        data_hash.set("image.data", buffer_handle)
        if self._schema_cache["sendGainMap"]:
            (
                buffer_handle,
                buffer_array,
            ) = self._shmem_buffer_gain_map.next_slot()
            self.gpu_runner.get_gain_map(
                output_order=self._schema_cache["dataFormat.outputAxisOrder"],
                out=buffer_array,
            )
            data_hash.set(
                "image.gainMap",
                buffer_handle,
            )
            data_hash.set("calngShmemPaths", ["image.data", "image.gainMap"])
        else:
            data_hash.set("calngShmemPaths", ["image.data"])

        data_hash.set("image.cellId", cell_table[:, np.newaxis])
        data_hash.set("image.pulseId", pulse_table[:, np.newaxis])

        self._write_output(data_hash, metadata)
        if do_generate_preview:
            if self._schema_cache["sendGainMap"]:
                self._write_combiner_previews(
                    (
                        ("preview.outputRaw", preview_raw),
                        ("preview.outputCorrected", preview_corrected),
                        ("preview.outputGain", preview_gain),
                    ),
                    train_id,
                    source,
                )
                # TODO: DRY
            else:
                self._write_combiner_previews(
                    (
                        ("preview.outputRaw", preview_raw),
                        ("preview.outputCorrected", preview_corrected),
                    ),
                    train_id,
                    source,
                )

    def _load_constant_to_gpu(self, constant, constant_data):
        # TODO: encode correction / constant dependencies in a clever way
        if constant is AgipdConstants.ThresholdsDark:
            field_name = "thresholding"  # TODO: (reverse) mapping, DRY
            if self.gain_mode is not AgipdGainMode.ADAPTIVE_GAIN:
                self.log.INFO("Loaded ThresholdsDark ignored due to fixed gain mode")
                return
            self.gpu_runner.load_thresholds(constant_data)
        elif constant is AgipdConstants.Offset:
            field_name = "offset"
            self.gpu_runner.load_offset_map(constant_data)
        elif constant is AgipdConstants.SlopesPC:
            field_name = "relGainPc"
            self.gpu_runner.load_rel_gain_pc_map(constant_data)
            if self._override_md_additional_offset is not None:
                self.gpu_runner.md_additional_offset_gpu.fill(
                    self._override_md_additional_offset
                )
        elif constant is AgipdConstants.SlopesFF:
            field_name = "relGainXray"
            self.gpu_runner.load_rel_gain_ff_map(constant_data)
        elif "BadPixels" in constant.name:
            field_name = "badPixels"
            self.gpu_runner.load_bad_pixels_map(
                constant_data, override_flags_to_use=self._override_bad_pixel_flags
            )

        # switch relevant correction on if it just now became available
        if not self.get(f"corrections.{field_name}.available"):
            # TODO: turn off again when flushing
            self.set(f"corrections.{field_name}.available", True)

        self._update_correction_flags()
        self.log_status_info(f"Done loading {constant.name} to GPU")

    def _update_pulse_filter(self, filter_string):
        """Called whenever the pulse filter changes, typically followed by
        _update_shapes"""

        if filter_string.strip() == "":
            new_filter = np.arange(self.get("dataFormat.memoryCells"), dtype=np.uint16)
        else:
            new_filter = np.array(eval(filter_string), dtype=np.uint16)
        assert np.max(new_filter) < self.get("dataFormat.memoryCells")
        self.pulse_filter = new_filter

    def _update_shapes(self):
        super()._update_shapes()
        # TODO: pack four pixels per byte
        if self._schema_cache["sendGainMap"]:
            if self._shmem_buffer_gain_map is None:
                buffer_name = self.getInstanceId() + ":dataOutput-gain"
                # try to match number of trains in image data buffer
                memory_budget = (
                    self.get("outputShmemBufferSize")
                    * 2 ** 30
                    // (
                        np.dtype(self.output_data_dtype).itemsize
                        // np.dtype(np.uint8).itemsize
                    )
                )
                self.log.INFO(f"Opening new shmem buffer for gain: {buffer_name}")
                self._shmem_buffer_gain_map = shmem_utils.ShmemCircularBuffer(
                    memory_budget,
                    self.output_data_shape,
                    np.uint8,
                    buffer_name,
                )
            else:
                self._shmem_buffer_gain_map.change_shape(self.output_data_shape)

    def _update_bad_pixel_selection(self):
        selection = 0
        for field in BadPixelValues:
            if self.get(f"corrections.badPixels.subsetToUse.{field.name}"):
                selection |= field
        self._override_bad_pixel_flags = selection

    def postReconfigure(self):
        super().postReconfigure()

        if self.get("corrections.relGainPc.overrideMdAdditionalOffset"):
            self._override_md_additional_offset = self.get(
                "corrections.relGainPc.mdAdditionalOffset"
            )
            self.gpu_runner.override_md_additional_offset(
                self._override_md_additional_offset
            )
        else:
            self._override_md_additional_offset = None

        if not hasattr(self, "_prereconfigure_update_hash"):
            return

        update = self._prereconfigure_update_hash

        if update.has("corrections.relGainXray.gGainValue"):
            self.gpu_runner.set_g_gain_value(
                self.get("corrections.relGainXray.gGainValue")
            )
            self._gpu_runner_init_args["g_gain_value"] = self.get(
                "corrections.relGainXray.gGainValue"
            )

        if update.has("corrections.badPixels.maskingValue"):
            self.bad_pixel_mask_value = eval(
                self.get("corrections.badPixels.maskingValue")
            )
            self.gpu_runner.set_bad_pixel_mask_value(self.bad_pixel_mask_value)
            self._gpu_runner_init_args[
                "bad_pixel_mask_value"
            ] = self.bad_pixel_mask_value

        if any(
            path.startswith("corrections.badPixels.subsetToUse")
            for path in update.getPaths()
        ):
            self.log_status_info("Updating bad pixel maps based on subset specified")
            if any(
                update.get(
                    f"corrections.badPixels.subsetToUse.{field.name}", default=False
                )
                for field in BadPixelValues
            ):
                self.log_status_info(
                    "Some fields reenabled, reloading cached bad pixel constants"
                )
                with self.calcat_friend.cached_constants_lock:
                    for (
                        constant,
                        data,
                    ) in self.calcat_friend.cached_constants.items():
                        if "BadPixels" in constant.name:
                            self._load_constant_to_gpu(constant, data)
            self._update_bad_pixel_selection()
            self.gpu_runner.override_bad_pixel_flags_to_use(
                self._override_bad_pixel_flags
            )
