#############################################################################
# Author: schmidtp
# Created on August 06, 2021, 05:06 PM
# Copyright (C) European XFEL GmbH Hamburg. All rights reserved.
#############################################################################

from asyncio import gather, wait_for, TimeoutError as AsyncTimeoutError
from collections import defaultdict
from collections.abc import Hashable
from datetime import datetime
from inspect import ismethod
from itertools import chain, repeat
from traceback import format_exc
from urllib.parse import urlparse
import json
import logging

from tornado.httpclient import AsyncHTTPClient, HTTPError
from tornado.platform.asyncio import AsyncIOMainLoop, to_asyncio_future
from pkg_resources import parse_version

from karabo.middlelayer import (
    KaraboError, Device, DeviceClientBase, Descriptor, Hash, Configurable,
    Slot, Node, Type,
    AccessMode, AccessLevel, Assignment, DaqPolicy, State, Unit,
    UInt16, UInt32, Bool, Double, Schema, String, VectorString, VectorHash,
    background, call, callNoWait, setNoWait, sleep, instantiate, slot, coslot,
    getDevice, getTopology, getConfiguration, getConfigurationFromPast,
    get_property)
from karabo.middlelayer_api.proxy import ProxyFactory

from karabo import version as karaboVersion
from ._version import version as deviceVersion
from . import scenes


'''
Device states:
 - INIT: When the device is starting up
 - ACTIVE: When the device is ready to manage a calibration pipeline
 - CHANGING: When the device is actively changing the pipeline configuration
 - ERROR: Recoverable error, only allows server restart
 - UNKNOWN: Unrecoverable error
'''


# Copied from karabo MDL source (location depending on version)
# Will be part of MDL's public API in 2.12
def get_instance_parent(instance):
    """Find the parent of the instance"""
    parent = instance
    while True:
        try:
            parent = next(iter(parent._parents))
        except StopIteration:
            break

    return parent


class ClassIdsNode(Configurable):
    correctionClass = String(
        displayedName='Correction class',
        description='Device class to use for corrections.',
        defaultValue='{}Correction',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    groupMatcherClass = String(
        displayedName='Group matcher class',
        description='Device class to use for matching the stream output of a '
                    'module group.',
        defaultValue='TrainMatcher',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    bridgeClass = String(
        displayedName='Bridge class',
        description='Device class to use for bridging the stream output out '
                    'of Karabo.',
        defaultValue='ShmemToZmq',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    previewMatcherClass = String(
        displayedName='Preview matcher class',
        description='Device class to use for matching the output of a preview '
                    'layer.',
        defaultValue='ModuleStacker',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    assemblerClass = String(
        displayedName='Assembler class',
        description='Device class to use for assembling the matched output of '
                    'a preview layer.',
        defaultValue='FemDataAssembler',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)


class DeviceIdsNode(Configurable):
    correctionSuffix = String(
        displayedName='Correction suffix',
        description='Suffix for correction device IDs. The formatting '
                    'placeholders \'virtualName\', \'index\' and \'group\' '
                    'may be used.',
        defaultValue='CORRECT{index:02d}_{virtualName}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    groupMatcherSuffix = String(
        displayedName='Group matcher suffix',
        description='Suffix for group matching device IDs. The formatting '
                    'placeholder \'group\' may be used.',
        defaultValue='MATCH_G{group}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    bridgeSuffix = String(
        displayedName='Bridge suffix',
        description='Suffix for group bridge device IDs. The formatting '
                    'placeholder \'group\' may be used.',
        defaultValue='BRIDGE_G{group}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    previewMatcherSuffix = String(
        displayedName='Preview matcher suffix',
        description='Suffix for preview layer matching device IDs. The '
                    'formatting placeholder \'layer\' may be used.',
        defaultValue='MATCH_{layer}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    assemblerSuffix = String(
        displayedName='Assembler suffix',
        description='Suffix for assembler device IDs. The formatting '
                    'placeholder \'layer\' may be used.',
        defaultValue='ASSEMBLE_{layer}',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)


class ModuleRow(Configurable):
    virtualName = String(
        displayedName='Virtual name')

    group = UInt32(
        displayedName='Group')

    aggregator = String(
        displayedName='Aggregator name')

    inputChannel = String(
        displayedName='Input channel')

    inputSource = String(
        displayedName='Input source')


class ModuleGroupRow(Configurable):
    group = UInt32(
        displayedName='Group')

    deviceServer = String(
        displayedName='Device server')

    withMatcher = Bool(
        displayedName='Matcher?')

    withBridge = Bool(
        displayedName='Bridge?')

    bridgePort = UInt16(
        displayedName='Bridge port',
        defaultValue=47000,
        minInc=1024,
        maxInc=65353)

    bridgePattern = String(
        displayedName='Bridge pattern',
        options=['PUSH', 'REP', 'PUBLISH'],
        defaultValue='PUSH')


class PreviewLayerRow(Configurable):
    layer = String(
        displayedName='Preview layer')

    outputPipeline = String(
        displayedName='Output pipeline')

    deviceServer = String(
        displayedName='Device server')


class DeviceServerRow(Configurable):
    deviceServer = String(
        displayedName='Device server')

    webserverHost = String(
        displayedName='Webserver host')


class WebserverApiNode(Configurable):
    statePollInterval = Double(
        displayedName='Status poll interval',
        description='Time between subsequent polls of the webserver API when '
                    'a device server is expected to reach a particular state.',
        defaultValue=0.5,
        minInc=0.2,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)

    downTimeout = Double(
        displayedName='Down timeout',
        description='Time to wait for a device server to stop gracefully '
                    'after a \'down\' command and until the \'kill\' command '
                    'is sent.',
        defaultValue=5.0,
        minInc=1.0,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)

    killTimeout = Double(
        displayedName='Kill timeout',
        description='Time to wait for a device server to stop after a '
                    '\'kill\' command and until the server is deemed '
                    'unrecoverable.',
        defaultValue=5.0,
        minInc=1.0,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)

    upTimeout = Double(
        displayedName='Up timeout',
        description='Time to wait for a device server to start successfully '
                    'after being down and until the server i deemed '
                    'broken.',
        defaultValue=5.0,
        minInc=1.0,
        unitSymbol=Unit.SECOND,
        accessMode=AccessMode.RECONFIGURABLE)


class InstantiationOptionsNode(Configurable):
    restoreMatcherSources = Bool(
        displayedName='Restore matcher sources',
        description='Attempt to retrieve and restore the last known '
                    'configuration for slow and fast sources of matcher '
                    'devices when the pipeline is instantiated.',
        defaultValue=False,
        accessMode=AccessMode.RECONFIGURABLE)

    autoActivateGroupBridges = Bool(
        displayedName='Activate bridges automatically',
        description='Whether to activate all group bridges immediately after '
                    'instantation.',
        defaultValue=False,
        accessMode=AccessMode.RECONFIGURABLE)

    autoActivateGroupMatchers = Bool(
        displayedName='Activate group matchers automatically',
        description='Whether to activate all group matchers immediately after '
                    'instantation.',
        defaultValue=False,
        accessMode=AccessMode.RECONFIGURABLE)


class ManagedKeysNode(Configurable):
    # Keys managed on detector DAQ devices.
    DAQ_KEYS = {'DataDispatcher.trainStride': 'daqTrainStride'}

    @UInt32(
        displayedName='DAQ train stride',
        unitSymbol=Unit.COUNT,
        defaultValue=5,
        allowedStates=[State.ACTIVE],
        minInc=1)
    async def daqTrainStride(self, value):
        self.daqTrainStride = value
        background(get_instance_parent(self)._set_on_daq(
            'DataDispatcher.trainStride', value))


class ManagedKeysCloneFactory(ProxyFactory):
    Proxy = ManagedKeysNode
    SubProxy = Configurable
    ProxyNode = Node
    node_factories = dict(Slot=Slot)


class CalibrationManager(DeviceClientBase, Device):
    __version__ = deviceVersion

    interfaces = VectorString(
        displayedName='Device interfaces',
        description='Interfaces implemented by this device.',
        defaultValue=(
            ['DeviceInstantiator']
            if parse_version(karaboVersion) >= parse_version('2.11')
            else []),
        accessMode=AccessMode.READONLY)

    availableScenes = VectorString(
        displayedName='Available scenes',
        displayType='Scenes',
        requiredAccessLevel=AccessLevel.OBSERVER,
        accessMode=AccessMode.READONLY,
        defaultValue=['overview'],
        daqPolicy=DaqPolicy.OMIT)

    @slot
    def requestScene(self, params):
        name = params.get('name', default='overview')
        if name == 'overview':
            # Assumes there are correction devices known to manager
            scene_data = scenes.manager_device_overview_scene(
                self.deviceId,
                self.getDeviceSchema(),
                self._correction_device_schema,
                self._correction_device_ids,
                self._domain_device_ids,
            )
            payload = Hash('success', True, 'name', name, 'data', scene_data)
            return Hash('type', 'deviceScene',
                        'origin', self.deviceId,
                        'payload', payload)

    detectorType = String(
        displayedName='Detector type',
        description='Type of the detector to manage.',
        options=['AGIPD', 'LPD', 'DSSC', 'Jungfrau', 'ePix100', 'pnCCD',
                 'FastCCD'],
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    detectorIdentifier = String(
        displayedName='Detector identifier',
        description='Name of this detector in CalCat and device ID domain '
                    'for the corresponding data aggregators, if applicable. ',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    classIds = Node(
        ClassIdsNode,
        displayedName='Class IDs',
        description='class IDs to instantiate for each role. A formatting '
                    'placeholer may be used to indicate the capitalized '
                    'detector type, e.g. Dssc or Jungfrau')

    deviceIds = Node(
        DeviceIdsNode,
        displayedName='Device IDs',
        description='Templates for the member field of the device IDs in the '
                    'pipeline, which is appended to the same device ID root '
                    '(DOMAIN/TYPE) as the manager instance itself.')

    modules = VectorHash(
        displayedName='Modules',
        description='Individual modules constituting this detector. A module '
                    'is the boundary on which correction constants are '
                    'calculated, and in general equivalent to PDUs. Only the '
                    'virtual name and group columns are obligatory, if left '
                    'blank the aggregator name is set to '
                    'f\'{detector_type.upper()}{index:02d}\', the input '
                    'channel is set to '
                    'f\'{detector_identifier}/DET/{index}CH0:output\' and the '
                    'input source '
                    'f\'{detector_identifier}/DET/{index}CH0:xtdf\'.',
        rows=ModuleRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)

    moduleGroups = VectorHash(
        displayedName='Module groups',
        rows=ModuleGroupRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)

    previewLayers = VectorHash(
        displayedName='Preview layers',
        rows=PreviewLayerRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)

    @VectorHash(
        displayedName='Device servers',
        rows=DeviceServerRow,
        accessMode=AccessMode.RECONFIGURABLE,
        assignment=Assignment.MANDATORY)
    async def deviceServers(self, value):
        self.deviceServers = value
        self._servers_changed = True

    geometryDevice = String(
        displayedName='Geometry device',
        description='[NYI] Device ID for a geometry device defining the '
                    'detector layout and module positions.',
        accessMode=AccessMode.INITONLY,
        assignment=Assignment.MANDATORY)

    calcatUrl = String(
        displayedName='CalCat URL',
        description='[NYI] URL to CalCat API to use for constant retrieval, '
                    'set by local secrets file',
        accessMode=AccessMode.READONLY)

    webserverApi = Node(
        WebserverApiNode,
        displayedName='Webserver API',
        description='Configurations for the webserver API to control device '
                    'servers.')

    instantiationOptions = Node(
        InstantiationOptionsNode,
        displayedName='Instantiation options',
        description='Optional flags controlling the pipeline instantiation.')

    doNotCompressEvents = Bool(
        requiredAccessLevel=AccessLevel.GOD,
        accessMode=AccessMode.READONLY,
        defaultValue=False,
        daqPolicy=DaqPolicy.OMIT)

    @Slot(
        displayedName='Restart servers',
        allowedStates=[State.ACTIVE, State.ERROR])
    async def restartServers(self):
        self.state = State.CHANGING
        background(self._restart_servers())

    @Slot(
        displayedName='Instantiate pipeline',
        allowedStates=[State.ACTIVE])
    async def startInstantiate(self):
        # Slot name is mandated by DeviceInstantiator interface.
        self.state = State.CHANGING
        background(self._instantiate_pipeline())

    @Slot(
        displayedName='Apply managed values',
        description='Set all managed keys to the values currently active on '
                    'the manager, replacing any manual change.',
        allowedStates=[State.ACTIVE])
    async def applyManagedValues(self):
        background(self._apply_managed_values())

    managed = Node(
        ManagedKeysNode,
        displayedName='Managed keys',
        description='Properties and slots managed on devices in the pipeline.')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Take the manager's device ID root (DOMAIN/TYPE) as the root
        # for all devices it manages.
        self._device_id_root = self.deviceId[:self.deviceId.rfind('/')].value

        # Concretized class ID for correction devices.
        self._correction_class_id = self.classIds.correctionClass.format(
            self.detectorType.value.lower().capitalize())

        # Set of data aggregators associated with the managed detector.
        self._daq_device_ids = set()

        # Set of devices in the same domain as the manager, i.e. having
        # the same device ID root.
        self._domain_device_ids = set()

        # Set of correction devices for the managed detector, i.e. being
        # in the same domain and having the specified class ID.
        self._correction_device_ids = set()

    async def onInitialization(self):
        self.state = State.INIT

        if parse_version(karaboVersion) < parse_version('2.11'):
            # Compatibility with 2.10 and before, where onInitialization
            # is awaited during device instantation.
            background(self._async_init())
        else:
            await self._async_init()

    @coslot
    async def slotInstanceNew(self, instance_id, info):
        # A new instance has appeared in the topology.

        await super().slotInstanceNew(instance_id, info)

        if info['type'] == 'device':
            self._check_new_device(instance_id, info['classId'])

    @slot
    def slotInstanceGone(self, instance_id, info):
        # An instance is gone from the topology.

        super().slotInstanceGone(instance_id, info)

        if info['type'] == 'device':
            self._daq_device_ids.discard(instance_id)
            self._domain_device_ids.discard(instance_id)
            self._correction_device_ids.discard(instance_id)

    async def _async_init(self):
        # Populate the device ID sets with what's out there right now.
        await self._check_topology()

        # Set-up Tornado.
        if hasattr(AsyncIOMainLoop, 'initialized'):
            if not AsyncIOMainLoop.initialized():
                AsyncIOMainLoop().install()

        self._http_client = AsyncHTTPClient()

        # Check device servers and initialize webserver access.
        await self._check_servers()

        # Inject schema for configuration of managed devices.
        await self._inject_managed_keys()

        if self.state == State.INIT:
            self._set_status('Calibration manager ready')
            self.state = State.ACTIVE

    def _check_new_device(self, device_id, class_id):
        if class_id == 'DataAggregator' and \
                device_id.startswith(self.detectorIdentifier.value):
            # This device is a data aggregator belonging to the detector
            # installation
            self._daq_device_ids.add(device_id)

        elif device_id.startswith(self._device_id_root):
            # This device lives under the same device ID root as this
            # manager instance.
            self._domain_device_ids.add(device_id)

            if class_id == self._correction_class_id:
                # This device is also a correction device.
                self._correction_device_ids.add(device_id)

    async def _check_topology(self):
        for i in range(10):
            try:
                devices = getTopology()['device']
            except AttributeError:
                # Not ready yet
                await sleep(1.0)
            else:
                if not devices:
                    await sleep(1.0)
                else:
                    break
        else:
            self._set_fatal(f'Topology not available after {i+1} tries')
            return

        self.logger.debug(f'Topology arrived after {i+1} tries')

        for device_id in devices:
            self._check_new_device(device_id, devices[device_id, 'classId'])

    async def _get_shared_keys(self, device_ids, keys):
        """Find the most common property values on devices."""

        key_values = defaultdict(list)

        for device_id in device_ids:
            try:
                config = await wait_for(getConfiguration(device_id),
                                        timeout=2.0)
            except AsyncTimeoutError:
                # Ignore this device if the configuration can no longer
                # be obtained.
                continue

            for key in keys:
                value = config[key]

                if isinstance(value, Hashable):
                    # Value must be hashable to determine the most
                    # common one below.
                    key_values[key].append(value)

        return {key: max(set(values), key=values.count) for key, values
                in key_values.items()}

    async def _inject_managed_keys(self):
        """Attempt to retrieve the correction device's schema and insert
        part of it as managed keys.
        """

        correction_device_servers = [
            server for _, server, _, _, _, _ in self.moduleGroups.value]

        up_corr_servers = await self._get_servers_in_state(
            'up', servers=correction_device_servers)

        if up_corr_servers:
            # At least one correction server is up.
            corr_server = next(iter(up_corr_servers))
        else:
            # None of the correction servers is up, try to start the
            # first one.
            corr_server = correction_device_servers[0]

            try:
                await wait_for(self._ensure_server_state(corr_server, 'up'),
                               timeout=self.webserverApi.upTimeout.value)
            except AsyncTimeoutError:
                self._set_fatal(f'Could not bring up correction device server '
                                f'`{corr_server}` for schema retrieval')
                return

        # Obtain the device schema from a correction device server.
        managed_schema, _, _ = await call(corr_server, 'slotGetClassSchema',
                                          self._correction_class_id)
        # saving this for later
        self._correction_device_schema = Schema()
        self._correction_device_schema.copy(managed_schema)

        if managed_schema.name != self._correction_class_id:
            self._set_fatal(
                f'Correction class ID `{self._correction_class_id}` not known '
                f'or loadable by device server `{corr_server}`')
            return

        # Collect the keys to be managed and build a nested hash
        # expressing its hierarchy, leafs are set to None.
        managed_keys = set(managed_schema.hash['managedKeys', 'defaultValue'])
        managed_tree = Hash(*chain.from_iterable(
            zip(managed_keys, repeat(None, len(managed_keys)))))
        managed_paths = set(managed_tree.paths())

        # Reduce the correction schema to the managed paths.
        managed_hash = managed_schema.hash
        for path in managed_hash.paths():
            if path not in managed_paths:
                del managed_hash[path]

        # Check for current values of managed keys on DAQ devices and
        # update schema correspondingly.
        try:
            if self._daq_device_ids:
                daq_vals = await self._get_shared_keys(
                    self._daq_device_ids, ManagedKeysNode.DAQ_KEYS.keys())

                for orig_key, managed_key in ManagedKeysNode.DAQ_KEYS.items():
                    managed_hash[managed_key, 'defaultValue'] = daq_vals[orig_key]
        except Exception as e:
            self.logger.warn(f'Error in getting defaults from DAQs: "{e}"')

        # Check for current values of managed keys on correction
        # devices and update schema correspondingly.
        corr_vals = await self._get_shared_keys(
            self._correction_device_ids, managed_keys)

        for key, value in corr_vals.items():
            managed_hash[key, 'defaultValue'] = value

        # Retrieve the attributes on the current managed node. The
        # original implementation of toSchemaAndAttrs in the Node's
        # superclass Descriptor is used to avoid Node-specific
        # attributes in the attributes that are not valid in the
        # property definition.
        # The value are then obtained from the Node object again since
        # enums are converted to their values by toSchemaAndAttrs, which
        # in turn is not valid for property definition.
        _, attrs = Descriptor.toSchemaAndAttrs(self.__class__.managed,
                                               None, None)
        managed_node_attrs = {key: getattr(self.__class__.managed, key)
                              for key in attrs.keys()}

        # Build a proxy from the managed schema, and create a new node
        # based on it with the original node attributes. This code is
        # heavily inspired by deviceClone.
        managed_node = Node(
            ManagedKeysCloneFactory.createProxy(managed_schema),
            **managed_node_attrs)

        # Walk the managed tree to and sanitize all descriptors to our
        # specifications.
        def _sanitize_node(parent, tree, prefix=''):
            for key, value in tree.items():
                # Fetch the descriptor class, not its instance!
                descr = getattr(parent.cls, key)

                full_key = f'{prefix}.{key}' if prefix else key

                if isinstance(descr, Node):
                    _sanitize_node(descr, value, full_key)

                elif isinstance(descr, Slot):
                    async def _managed_slot_called(parent, fk=full_key):
                        background(self._call_on_corrections(fk))

                    _managed_slot_called.__name__ = f'managed.{full_key}'
                    descr.__call__(_managed_slot_called)

                    # Managed slots can only be called in the ACTIVE
                    # state.
                    descr.allowedStates = [State.ACTIVE]

                elif isinstance(descr, Type):
                    # Regular property.

                    if descr.accessMode == AccessMode.RECONFIGURABLE:
                        # Add a callback only if the original descriptor
                        # is reconfigurable.

                        async def _managed_prop_changed(parent, v, k=key,
                                                        fk=full_key):
                            setattr(parent, k, v)

                            if self.state != State.INIT:
                                # Do not propagate updates during injection.
                                background(self._set_on_corrections(fk, v))

                        descr.__call__(_managed_prop_changed)

                    # Managed properties are always optional,
                    # reconfigurable and may only be changed in the
                    # ACTIVE state.
                    descr.assignment = Assignment.OPTIONAL
                    descr.accessMode = AccessMode.RECONFIGURABLE
                    descr.allowedStates = [State.ACTIVE]

                else:
                    self.logger.warn(f'Encountered unknown descriptor type '
                                     f'{type(descr)}')

        _sanitize_node(managed_node, managed_tree)

        # Inject the newly prepared node for managed keys.
        self.__class__.managed = managed_node
        await self.publishInjectedParameters()
        self._managed_keys = managed_keys

        self.logger.debug('Managed schema injected')

    def _set_status(self, text, level=logging.INFO):
        """Add and log a status message.

        Suppresses throttling from the gui server.
        """

        self.status = text
        self.doNotCompressEvents = not self.doNotCompressEvents
        self.logger.log(level, text)

    def _set_error(self, text):
        """Set the device into error state and log an error message."""

        self.state = State.ERROR
        self._set_status(text, level=logging.ERROR)

    def _set_fatal(self, text):
        """Set the device into unknown state and log an error message."""

        self.state = State.UNKNOWN
        self._set_status(text, level=logging.CRITICAL)

    def _set_exception(self, text, e):
        """Set the device into error upon an exception."""

        return self._set_error(f'{text}\n{format_exc()[:-1]}')

    def _get_server_api_url(self, name):
        """Get API URL for a device server."""

        return '{host}/api/servers/{api_name}.json'.format(
            host=self._server_hosts[name],
            api_name=self._server_api_names[name])

    async def _get_server_info(self, name):
        """Get info for a device server.

        Args:
            name (str): Device server (Karabo) name.

        Returns:
            (dict) Information returned by webserver for this server.
        """

        try:
            reply = await to_asyncio_future(self._http_client.fetch(
                self._get_server_api_url(name), method='GET'))
        except (ConnectionError, HTTPError) as e:
            raise RuntimeError(
                f'Failed retrieving server info of {name} on '
                f'{urlparse(self._server_hosts[name]).hostname}: {e}')

        body = json.loads(reply.body)

        if not body['success']:
            raise RuntimeError(f'Request to retrieve server info for {name} '
                               f'failed')

        self.logger.debug(f'Obtained server info for {name}: '
                          f'{body["servers"][0]}')

        return body['servers'][0]

    async def _get_servers_in_state(self, state, servers=None):
        """List all servers in a particular state.

        Args:
            state (str): State to filter for, e.g. 'up' or 'down'.

        Returns:
            (list[str]) List of server names found in the passed state.
        """

        servers = await gather(*[self._get_server_info(name)
                                 for name in self._server_hosts.keys()
                                 if servers is None or name in servers])

        return {server['karabo_name'] for server in servers
                if server['status'].startswith(state)}

    async def _ensure_server_state(self, name, state, command=None):
        """Sets and waits for a a server state to be reached.

        Args:
            name (str): Device server (Karabo) name.
            state (str): State to set and wait for, e.g. 'up' or 'down'.
            command (str, optional): Command to use for reaching the
                state, same as state of omitted.
        """

        if command is None:
            command = state

        try:
            reply = await to_asyncio_future(self._http_client.fetch(
                self._get_server_api_url(name), method='PUT',
                body=json.dumps({'server': {'command': command}})))
        except (ConnectionError, HTTPError) as e:
            raise RuntimeError(
                f'Failed sending `{command}` to {name} on '
                f'{urlparse(self._server_hosts[name]).hostname}: {e}')

        if not json.loads(reply.body)['success']:
            raise RuntimeError('Command `{command}` for {name} failed')

        while True:
            await sleep(self.webserverApi.statePollInterval.value)
            info = await self._get_server_info(name)

            if info['status'].startswith(state):
                return

    async def _check_servers(self):
        """Validate device server configuration."""

        if not self._servers_changed:
            return True

        # Mapping of servers to device servers.
        self._server_hosts = {server: host for server, host
                              in self.deviceServers.value}

        # Mapping of "Karabo names" (with capitalization and slashes)
        # to "API names" (all lower case and only underscores).
        self._server_api_names = {}

        # Build a mapping of hosts to server names.
        hosts = defaultdict(set)
        for name, host in self._server_hosts.items():
            hosts[host].add(name)

        # List of errors found during device server check.
        err = []

        # Query all servers on each host and check whether the ones we
        # need are in there, and obtain their API names.
        for host, req_names in hosts.items():
            # Retrieve hostname for nice error messages.
            hostname = urlparse(self._server_hosts[name]).hostname

            try:
                reply = await to_asyncio_future(self._http_client.fetch(
                    f'{host}/api/servers.json', method='GET'))
            except (ConnectionError, HTTPError) as e:
                err.append(f'- {e.__class__.__name__} when retrieving server '
                           f'list on {hostname}: {e}')
                continue

            body = json.loads(reply.body)

            if not body['success']:
                err.append(f'- Request to retrieve server list failed on '
                           f'{hostname}')
                continue

            servers = {s['karabo_name']: s for s in body['servers']
                       if s['karabo_name'] in req_names}

            if len(servers) != len(req_names):
                err.append(f'- Device servers missing on {hostname}: ' +
                           ', '.join(req_names - servers.keys()))

            servers_in_error = {s['karabo_name'] for s in servers.values()
                                if 'error' in s['status']}

            if servers_in_error:
                err.append(f'- Device servers on {hostname} are in error state'
                           + ', '.join(servers_in_error))
                continue

            servers_disabled = {s['karabo_name'] for s in servers.values()
                                if not s['control_allowed']}

            if servers_disabled:
                err.append(f'- Device servers on {hostname} not controllable '
                           f'via webserver: ' + ', '.join(servers_disabled))
                continue

            for name, info in servers.items():
                self._server_api_names[name] = info['name']

        if err:
            self._set_error('One or more device server problems were '
                            'detected:\n' + '\n'.join(err))
            return False
        else:
            return True

    async def _restart_servers(self):
        """Restart all managed device servers."""

        if not await self._check_servers():
            return

        try:
            # Find all servers in need of being brought down.
            up_names = await self._get_servers_in_state('up')

            # Bring down all servers which are up.
            if up_names:
                try:
                    await wait_for(gather(
                        *[self._ensure_server_state(name, 'down')
                          for name in up_names]),
                        timeout=self.webserverApi.downTimeout.value)
                except AsyncTimeoutError:
                    # Narrow down to the list of servers STILL up.
                    up_names = await self._get_servers_in_state('up')
                else:
                    up_names.clear()

            # If some servers are still up, go for the kill.
            if up_names:
                self.logger.warn('Some servers still up after waiting for '
                                 f'downTimeout: {up_names}')

                try:
                    await wait_for(gather(
                        *[self._ensure_server_state(name, 'down', 'kill')
                          for name in up_names]),
                        timeout=self.webserverApi.killTimeout.value)
                except AsyncTimeoutError:
                    # All hope is lost.
                    up_names = await self._get_servers_in_state('up')
                    return self._set_error(
                        'One or more device servers could not be brought '
                        'down, even via kill:' + ', '.join(up_names))

            # Bring all servers up again.
            try:
                await wait_for(gather(
                    *[self._ensure_server_state(name, 'up')
                      for name in self._server_hosts.keys()]),
                    timeout=self.webserverApi.upTimeout.value)
            except AsyncTimeoutError:
                down_names = await self._get_servers_in_state('down')
                return self._set_error('One or more device servers could not '
                                       'brought up: ' + ', '.join(down_names))

        except RuntimeError as e:
            return self._set_exception(
                'Request unexpectedly failed during restart procedure', e)

        # Wait a moment for good measure until the servers are
        # actually up. If we try to instantiate to quickly after the
        # servers reported running, instantiate() blocks forever.
        await sleep(5.0)

        if self._domain_device_ids:
            # If there are still devices in this set, log a warning.
            self.logger.warn('Some devices left in manager domain after '
                             'device servers have been restarted: '
                             + ', '.join(self._domain_device_ids))

        self._set_status('All device servers restarted')
        self.state = State.ACTIVE

    async def _instantiate_device(self, server, class_id, device_id, config):
        """Instantiate a single device.

        Small wrapper around karabo.middlelayer.instantiate for error
        handling and logging.
        """

        if device_id in self._domain_device_ids:
            self.logger.debug(f'Skipped instantiation of already existing '
                              f'device {device_id}')
            return True

        try:
            msg = await wait_for(instantiate(
                server, class_id, device_id, config), 5.0)
        except AsyncTimeoutError:
            self._set_error(f'Instantiation timeout on {device_id}')
            return False
        except KaraboError as e:
            self._set_exception(f'Instantiation error on {device_id}', e)
            return False

        self.logger.debug(f'Instantation result for {device_id}: {msg}')
        return True

    async def _instantiate_pipeline(self):
        """Instantiate all managed devices."""

        if not await self._check_servers():
            return

        # Make sure all servers are up.
        try:
            up_servers = await self._get_servers_in_state('up')

            if up_servers != self._server_api_names.keys():
                self.state = State.ACTIVE
                return self._set_status('One or more device servers are not '
                                        'up, restart servers first.')
        except RuntimeError as e:
            return self._set_error('Request unexpectedly failed while '
                                   'checking device server state', e)

        # Class and device ID templates per role.
        class_ids = {}
        device_id_templates = {}

        class_args = (self.detectorType.value.lower().capitalize(),)
        for role in ['correction', 'groupMatcher', 'bridge', 'previewMatcher',
                     'assembler']:
            class_ids[role] = getattr(
                self.classIds, f'{role}Class').value.format(*class_args)
            device_id_templates[role] = f'{self._device_id_root}/' + \
                getattr(self.deviceIds, f'{role}Suffix')

        # Servers by group and layer.
        server_by_group = {group: server for group, server, _, _, _, _
                           in self.moduleGroups.value}
        server_by_layer = {layer: server for layer, _, server
                           in self.previewLayers.value}

        all_req_servers = set(server_by_group.values()).union(
            server_by_layer.values())

        if all_req_servers != up_servers:
            return self._set_error('One or more device servers are not '
                                   'listed in the device servers '
                                   'configuration')

        # Instantiate modules.
        modules_by_group = defaultdict(list)
        correct_device_id_by_module = {}
        input_source_by_module = {}

        for index, row in enumerate(self.modules.value):
            vname, group, aggregator, input_channel, input_source = row

            modules_by_group[group].append(vname)
            device_id = device_id_templates['correction'].format(
                virtualName=vname, index=index, group=group)
            correct_device_id_by_module[vname] = device_id

            if not aggregator:
                aggregator = f'{self.detectorType.upper()}{index:02d}'

            daq_device_id = f'{self.detectorIdentifier}/DET/{index}CH0'

            if not input_channel:
                input_channel = f'{daq_device_id}:output'

            if not input_source:
                input_source = f'{daq_device_id}:xtdf'

            input_source_by_module[vname] = input_source

            config = Hash()

            config['constantParameters.detectorName'] = self.detectorIdentifier.value
            config['constantParameters.karaboDa'] = aggregator
            config['dataInput.connectedOutputChannels'] = [input_channel]
            config['fastSources'] = [input_source]

            # Add managed keys.
            for key in self._managed_keys:
                value = get_property(self, f'managed.{key}')

                if not ismethod(value):
                    config[key] = value

            if not await self._instantiate_device(
                server_by_group[group], class_ids['correction'], device_id,
                config
            ):
                return

        # Instantiate group matchers and bridges.
        for row in self.moduleGroups.value:
            group, server, with_matcher, with_bridge, bridge_port, \
                bridge_pattern = row

            # Group matcher, if applicable.
            if with_matcher:
                matcher_device_id = device_id_templates['groupMatcher'].format(
                    group=group)

                config = Hash()
                config['channels'] = [
                    f'{correct_device_id_by_module[vname]}:dataOutput'
                    for vname in modules_by_group[group]]
                config['fastSources'] = [
                    Hash('fsSelect', True,
                         'fsSource', input_source_by_module[vname])
                    for vname in modules_by_group[group]]

                if self.instantiationOptions.restoreMatcherSources:
                    try:
                        old_config = await getConfigurationFromPast(
                            matcher_device_id, datetime.now().isoformat())
                    except KaraboError:
                        pass  # Ignore configuration on error
                    else:
                        config['channels'] = old_config['channels']
                        config['slowSources'] = old_config['slowSources']
                        config['fastSources'] = old_config['fastSources']

                if not await self._instantiate_device(
                    server, class_ids['groupMatcher'],
                    matcher_device_id, config
                ):
                    return
                elif self.instantiationOptions.autoActivateGroupMatchers:
                    async def _activate_matcher(device_id):
                        with await getDevice(device_id) as device:
                            await sleep(3)
                            if device.state == State.PASSIVE:
                                await device.start()

                    background(_activate_matcher(matcher_device_id))

            # Group bridge, if applicable.
            if with_bridge:
                bridge_device_id = device_id_templates['bridge'].format(
                    group=group)

                config = Hash()
                config['outputsConfig'] = [Hash(
                    'pattern', bridge_pattern, 'hwm', 1, 'port', bridge_port)]

                config['input.connectedOutputChannels'] = [
                    f'{matcher_device_id}:output']

                if not await self._instantiate_device(
                    server, class_ids['bridge'], bridge_device_id, config
                ):
                    return
                elif self.instantiationOptions.autoActivateGroupBridges:
                    # Delay the slot a bit since it will get lost during
                    # instantation.

                    async def _activate_bridge(device_id):
                        with await getDevice(device_id) as device:
                            await sleep(3)
                            if device.state == State.PASSIVE:
                                await device.activate()

                    background(_activate_bridge(bridge_device_id))

        # Instantiate preview layer matchers and assemblers.
        for layer, output_pipeline, server in self.previewLayers.value:
            # Preview matcher.
            matcher_device_id = device_id_templates['previewMatcher'].format(
                layer=layer)

            config = Hash()
            config['channels'] = [
                f'{device_id}:{output_pipeline}'
                for device_id in correct_device_id_by_module.values()]
            config['fastSources'] = [
                Hash('fsSelect', True,
                     'fsSource',
                     f'{input_source_by_module[virtual_id]}')
                for (virtual_id, device_id)
                in correct_device_id_by_module.items()]
            config['pathToStack'] = 'data.adc'

            if not await self._instantiate_device(
                server, class_ids['previewMatcher'], matcher_device_id, config
            ):
                return

            # Preview assembler.
            assembler_device_id = device_id_templates['assembler'].format(
                layer=layer)

            config = Hash()
            config['input.connectedOutputChannels'] = [
                f'{matcher_device_id}:output']
            config['modules'] = [
                Hash('source', input_source_by_module.get('Q1M1', ''),
                     'offX', 474, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q1M2', ''),
                     'offX', 316, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q1M3', ''),
                     'offX', 158, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q1M4', ''),
                     'offX', 0, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q2M1', ''),
                     'offX', 1136, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q2M2', ''),
                     'offX', 978, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q2M3', ''),
                     'offX', 820, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q2M4', ''),
                     'offX', 662, 'offY', 612, 'rot', 90),
                Hash('source', input_source_by_module.get('Q3M1', ''),
                     'offX', 712, 'offY', 0, 'rot', 270),
                Hash('source', input_source_by_module.get('Q3M2', ''),
                     'offX', 870, 'offY', 0, 'rot', 270),
                Hash('source', input_source_by_module.get('Q3M3', ''),
                     'offX', 1028, 'offY', 0, 'rot', 270),
                Hash('source', input_source_by_module.get('Q3M4', ''),
                     'offX', 1186, 'offY', 0, 'rot', 270),
                Hash('source', input_source_by_module.get('Q4M1', ''),
                     'offX', 50, 'offY', 0, 'rot', 270),
                Hash('source', input_source_by_module.get('Q4M2', ''),
                     'offX', 208, 'offY', 0, 'rot', 270),
                Hash('source', input_source_by_module.get('Q4M3', ''),
                     'offX', 366, 'offY', 0, 'rot', 270),
                Hash('source', input_source_by_module.get('Q4M4', ''),
                     'offX', 524, 'offY', 0, 'rot', 270),
            ]
            config['pathsToCombine'] = ['data.adc']
            config['trainIdPath'] = 'image.trainId'
            config['pulseIdPath'] = 'image.pulseId'
            config['preview.enablePreview'] = True
            config['preview.pathToPreview'] = 'data.adc'
            config['preview.downSample'] = 2
            config['badpixelPath'] = 'image.bad_pixels'
            config['rotated90Grad'] = True

            if not await self._instantiate_device(
                server, class_ids['assembler'], assembler_device_id, config
            ):
                return

        self._set_status('All devices instantiated')
        self.state = State.ACTIVE

    async def _apply_managed_values(self):
        """Apply all managed keys to local values."""

        for daq_key, local_key in ManagedKeysNode.DAQ_KEYS.items():
            await self._set_on_daq(
                daq_key, get_property(self, f'managed.{local_key}'))

        for key in self._managed_keys:
            value = get_property(self, f'managed.{key}')
            if not ismethod(value):
                await self._set_on_corrections(key, value)

    def _call(self, device_ids, slot, *args):
        """Call the same slot on a list of devices."""

        for device_id in device_ids:
            callNoWait(device_id, slot, *args)
            self.logger.debug(f'Called {device_id}.{slot}{args}')

    async def _call_on_corrections(self, slot, *args):
        """Call a slot on all correction devices."""

        if self._correction_device_ids:
            self._call(self._correction_device_ids, slot, *args)
            self.logger.info(f'Called <CORR>.{slot}{args}')

    def _set(self, device_ids, key, value):
        """Set the same property on a list of devices."""

        for device_id in device_ids:
            setNoWait(device_id, key, value)
            self.logger.debug(f'Set {device_id}.{key} to {value}')

    async def _set_on_daq(self, key, value):
        """Set a property on all DAQ devices."""

        if self._daq_device_ids:
            self._set(self._daq_device_ids, key, value)
            self.logger.info(f'Set <DAQ>.{key} to {value}')

    async def _set_on_corrections(self, key, value):
        """Set a property on all correction devices."""

        if self._correction_device_ids:
            self._set(self._correction_device_ids, key, value)
            self.logger.info(f'Set <CORR>.{key} to {value}')
