import enum

import cupy
import numpy as np

from . import base_gpu, utils


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    THRESHOLD = 1
    OFFSET = 2
    BLSHIFT = 4
    REL_GAIN_PC = 8
    REL_GAIN_XRAY = 16
    BPMASK = 32


# from pycalibration's enum.py
class AgipdGainMode(enum.IntEnum):
    ADAPTIVE_GAIN = 0
    FIXED_HIGH_GAIN = 1
    FIXED_MEDIUM_GAIN = 2
    FIXED_LOW_GAIN = 3


class AgipdGpuRunner(base_gpu.BaseGpuRunner):
    _kernel_source_filename = "agipd_gpu_kernels.cpp"
    _corrected_axis_order = "cxy"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells,
        input_data_dtype=cupy.uint16,
        output_data_dtype=cupy.float32,
        bad_pixel_mask_value=cupy.nan,
        gain_mode=AgipdGainMode.ADAPTIVE_GAIN,
        g_gain_value=1,
        output_gain_map=False,
    ):
        self.gain_mode = gain_mode
        if self.gain_mode is AgipdGainMode.ADAPTIVE_GAIN:
            self.default_gain = cupy.uint8(gain_mode)
        else:
            self.default_gain = cupy.uint8(gain_mode - 1)
        self.input_shape = (memory_cells, 2, pixels_x, pixels_y)
        self.processed_shape = (memory_cells, pixels_x, pixels_y)
        self.output_gain_map = output_gain_map
        super().__init__(
            pixels_x,
            pixels_y,
            memory_cells,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        if self.output_gain_map:
            self.gain_map_gpu = cupy.empty(self.processed_shape, dtype=cupy.uint8)
            self.preview_gain = np.empty(self.preview_shape, dtype=np.float32)
        else:
            # TODO: don't even set in this case
            self.gain_map_gpu = cupy.empty(0, dtype=cupy.uint8)

        self.map_shape = (self.constant_memory_cells, self.pixels_x, self.pixels_y)
        self.gm_map_shape = self.map_shape + (3,)  # for gain-mapped constants
        self.threshold_map_shape = self.map_shape + (2,)
        # constants
        self.gain_thresholds_gpu = cupy.empty(
            self.threshold_map_shape, dtype=cupy.float32
        )
        self.offset_map_gpu = cupy.zeros(self.gm_map_shape, dtype=cupy.float32)
        self.rel_gain_pc_map_gpu = cupy.ones(self.gm_map_shape, dtype=cupy.float32)
        # not gm_map_shape because it only applies to medium gain pixels
        self.md_additional_offset_gpu = cupy.zeros(self.map_shape, dtype=cupy.float32)
        self.rel_gain_xray_map_gpu = cupy.ones(self.map_shape, dtype=cupy.float32)
        self.bad_pixel_map_gpu = cupy.zeros(self.gm_map_shape, dtype=cupy.uint32)
        self.set_bad_pixel_mask_value(bad_pixel_mask_value)
        self.set_g_gain_value(g_gain_value)

        self.update_block_size((1, 1, 64))

    def compute_preview_gain(self, preview_index):
        assert self.output_gain_map
        # TODO: abstract most of this in base_gpu to DRY
        if preview_index < -4:
            raise ValueError(f"No statistic with code {preview_index} defined")
        elif preview_index >= self.memory_cells:
            raise ValueError(f"Memory cell index {preview_index} out of range")

        if preview_index >= 0:
            self.gain_map_gpu[preview_index].astype(np.float32).get(
                out=self.preview_gain
            )
        elif preview_index == -1:
            # TODO: confirm that max is pixel and not integrated intensity
            # separate from next case because dtype not applicable here
            cupy.max(self.gain_map_gpu, axis=0).astype(cupy.float32).get(
                out=self.preview_gain
            )
        elif preview_index in (-2, -3, -4):
            stat_fun = {-1: cupy.max, -2: cupy.mean, -3: cupy.sum, -4: cupy.std}[
                preview_index
            ]
            stat_fun(self.gain_map_gpu, axis=0, dtype=cupy.float32).get(
                out=self.preview_gain
            )
        return self.preview_gain

    def _preview_preprocess_raw(self):
        return self.input_data_gpu[:, 0]

    def _preview_preprocess_corr(self):
        return self.processed_data_gpu

    def load_thresholds(self, threshold_map):
        # shape: y, x, memory cell, thresholds and gain values
        # note: the gain values are something like means used to derive thresholds
        self.gain_thresholds_gpu.set(
            np.transpose(threshold_map[..., :2], (2, 1, 0, 3)).astype(np.float32)
        )

    def load_offset_map(self, offset_map):
        # shape: y, x, memory cell, gain stage
        self.offset_map_gpu.set(
            np.transpose(offset_map, (2, 1, 0, 3)).astype(np.float32)
        )

    def load_rel_gain_pc_map(self, slopes_pc_map, override_md_additional_offset=None):
        # pc has funny shape (11, 352, 128, 512) from file
        # this is (fi, memory cell, y, x)
        slopes_pc_map = slopes_pc_map.astype(np.float32)
        # the following may contain NaNs, though...
        hg_slope = slopes_pc_map[0]
        hg_intercept = slopes_pc_map[1]
        mg_slope = slopes_pc_map[3]
        mg_intercept = slopes_pc_map[4]
        # TODO: remove sanitization (should happen in constant preparation notebook)
        # from agipdlib.py: replace NaN with median (per memory cell)
        # note: suffixes in agipdlib are "_m" and "_l", should probably be "_I"
        for naughty_array in (hg_slope, hg_intercept, mg_slope, mg_intercept):
            medians = np.nanmedian(naughty_array, axis=(1, 2))
            nan_bool = np.isnan(naughty_array)
            nan_cell, _, _ = np.where(nan_bool)
            naughty_array[nan_bool] = medians[nan_cell]

            too_low_bool = naughty_array < 0.8 * medians[:, np.newaxis, np.newaxis]
            too_low_cell, _, _ = np.where(too_low_bool)
            naughty_array[too_low_bool] = medians[too_low_cell]

            too_high_bool = naughty_array > 1.2 * medians[:, np.newaxis, np.newaxis]
            too_high_cell, _, _ = np.where(too_high_bool)
            naughty_array[too_high_bool] = medians[too_high_cell]

        frac_hg_mg = hg_slope / mg_slope
        rel_gain_map = np.ones(
            (3, self.constant_memory_cells, self.pixels_y, self.pixels_x),
            dtype=np.float32,
        )
        rel_gain_map[1] = rel_gain_map[0] * frac_hg_mg
        rel_gain_map[2] = rel_gain_map[1] * 4.48
        self.rel_gain_pc_map_gpu.set(np.transpose(rel_gain_map, (1, 3, 2, 0)))
        if override_md_additional_offset is None:
            md_additional_offset = (hg_intercept - mg_intercept * frac_hg_mg).astype(
                np.float32
            )
            self.md_additional_offset_gpu.set(
                np.transpose(md_additional_offset, (0, 2, 1))
            )
        else:
            self.override_md_additional_offset(override_md_additional_offset)

    def override_md_additional_offset(self, override_value):
        self.md_additional_offset_gpu.fill(override_value)

    def load_rel_gain_ff_map(self, slopes_ff_map):
        # constant shape: y, x, memory cell
        if slopes_ff_map.shape[2] == 2:
            # TODO: remove support for old format
            # old format, is per pixel only (shape is y, x, 2)
            # note: we should not support this in online
            slopes_ff_map = np.broadcast_to(
                slopes_ff_map[..., 0][..., np.newaxis],
                (self.pixels_y, self.pixels_x, self.constant_memory_cells),
            )
        self.rel_gain_xray_map_gpu.set(np.transpose(slopes_ff_map).astype(np.float32))

    def set_g_gain_value(self, override_value):
        self.g_gain_value = cupy.float32(override_value)

    def load_bad_pixels_map(self, bad_pixels_map, override_flags_to_use=None):
        print(f"Loading bad pixels with shape: {bad_pixels_map.shape}")
        # will simply OR with already loaded, does not take into account which ones
        # TODO: inquire what "mask for double size pixels" means
        if len(bad_pixels_map.shape) == 3:
            if bad_pixels_map.shape == (
                self.pixels_y,
                self.pixels_x,
                self.constant_memory_cells,
            ):
                # BadPixelsFF is not per gain stage - broadcasting along gain dimension
                self.bad_pixel_map_gpu |= cupy.asarray(
                    np.broadcast_to(
                        np.transpose(bad_pixels_map)[..., np.newaxis],
                        self.gm_map_shape,
                    ),
                    dtype=np.uint32,
                )
            elif bad_pixels_map.shape == (
                self.constant_memory_cells,
                self.pixels_y,
                self.pixels_x,
            ):
                # oh, can also be old bad pixels pc?
                self.bad_pixel_map_gpu |= cupy.asarray(
                    np.broadcast_to(
                        np.transpose(bad_pixels_map, (0, 2, 1))[..., np.newaxis],
                        self.gm_map_shape,
                    ),
                    dtype=np.uint32,
                )
            else:
                raise ValueError(
                    f"What in the world is this shape? {bad_pixels_map.shape}"
                )
        else:
            self.bad_pixel_map_gpu |= cupy.asarray(
                np.transpose(bad_pixels_map, (2, 1, 0, 3)), dtype=np.uint32
            )

        if override_flags_to_use is not None:
            self.override_bad_pixel_flags_to_use(override_flags_to_use)

    def override_bad_pixel_flags_to_use(self, override_value):
        self.bad_pixel_map_gpu &= cupy.uint32(override_value)

    def set_bad_pixel_mask_value(self, mask_value):
        self.bad_pixel_mask_value = cupy.float32(mask_value)

    def flush_buffers(self):
        self.offset_map_gpu.fill(0)
        self.rel_gain_pc_map_gpu.fill(1)
        self.md_additional_offset_gpu.fill(0)
        self.rel_gain_xray_map_gpu.fill(1)
        self.bad_pixel_map_gpu.fill(0)

    # TODO: baseline shift

    def correct(self, flags):
        if flags & CorrectionFlags.BLSHIFT:
            raise NotImplementedError("Baseline shift not implemented yet")
        if self.gain_mode is not AgipdGainMode.ADAPTIVE_GAIN and (
            flags & CorrectionFlags.THRESHOLD
        ):
            raise ValueError("Cannot do gain thresholding in fixed gain mode")

        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                cupy.uint8(flags),
                self.default_gain,
                self.gain_thresholds_gpu,
                self.offset_map_gpu,
                self.rel_gain_pc_map_gpu,
                self.md_additional_offset_gpu,
                self.rel_gain_xray_map_gpu,
                self.g_gain_value,
                self.bad_pixel_map_gpu,
                self.bad_pixel_mask_value,
                self.gain_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def get_gain_map(self, output_order, out=None):
        assert self.output_gain_map
        return cupy.ascontiguousarray(
            cupy.transpose(
                self.gain_map_gpu,
                utils.transpose_order(self._corrected_axis_order, output_order),
            )
        ).get(out=out)

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "data_memory_cells": self.memory_cells,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                "output_gain_map": self.output_gain_map,
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")


class BadPixelValues(enum.IntFlag):
    """The European XFEL Bad Pixel Encoding

    Straight from pycalibration's enum.py"""

    OFFSET_OUT_OF_THRESHOLD = 2 ** 0
    NOISE_OUT_OF_THRESHOLD = 2 ** 1
    OFFSET_NOISE_EVAL_ERROR = 2 ** 2
    NO_DARK_DATA = 2 ** 3
    CI_GAIN_OUT_OF_THRESHOLD = 2 ** 4
    CI_LINEAR_DEVIATION = 2 ** 5
    CI_EVAL_ERROR = 2 ** 6
    FF_GAIN_EVAL_ERROR = 2 ** 7
    FF_GAIN_DEVIATION = 2 ** 8
    FF_NO_ENTRIES = 2 ** 9
    CI2_EVAL_ERROR = 2 ** 10
    VALUE_IS_NAN = 2 ** 11
    VALUE_OUT_OF_RANGE = 2 ** 12
    GAIN_THRESHOLDING_ERROR = 2 ** 13
    DATA_STD_IS_ZERO = 2 ** 14
    ASIC_STD_BELOW_NOISE = 2 ** 15
    INTERPOLATED = 2 ** 16
    NOISY_ADC = 2 ** 17
    OVERSCAN = 2 ** 18
    NON_SENSITIVE = 2 ** 19
    NON_LIN_RESPONSE_REGION = 2 ** 20
