import enum

import karathon
from karabo.common.scenemodel.api import (
    CheckBoxModel,
    ColorBoolModel,
    DeviceSceneLinkModel,
    DisplayCommandModel,
    DisplayLabelModel,
    DisplayStateColorModel,
    DisplayTextLogModel,
    DoubleLineEditModel,
    EvaluatorModel,
    IntLineEditModel,
    LabelModel,
    LineEditModel,
    RectangleModel,
    SceneModel,
    SceneTargetWindow,
    write_scene,
)

BASE_INC = 25
NARROW_INC = 20
PADDING = 5
RECONFIGURABLE = 4  # TODO: look up proper enum

_type_to_line_editable = {
    "DOUBLE": (DoubleLineEditModel, {}),
    "FLOAT": (DoubleLineEditModel, {}),
    "INT32": (IntLineEditModel, {}),
    "UINT32": (IntLineEditModel, {}),
    "INT64": (IntLineEditModel, {}),
    "UINT64": (IntLineEditModel, {}),
    "STRING": (LineEditModel, {"klass": "EditableLineEdit"}),
}


def safe_render(obj, x, y):
    if hasattr(obj, "render"):
        return obj.render(x, y)
    else:
        obj.x = x
        obj.y = y
        return [obj]


class Align(enum.Enum):
    CENTER = enum.auto()
    TOP = enum.auto()
    BOTTOM = enum.auto()
    LEFT = enum.auto()
    RIGHT = enum.auto()


def titled(title, width=8 * NARROW_INC):
    def actual_decorator(component_class):
        class new_class(component_class):
            def render(self, x, y, *args, **kwargs):
                return [
                    LabelModel(
                        frame_width=1,
                        text=title,
                        width=width,
                        height=NARROW_INC,
                        x=x,
                        y=y,
                    )
                ] + component_class.render(self, x, y + NARROW_INC, *args, **kwargs)

            @property
            def width(self):
                return max(component_class.width.fget(self), width)

            @property
            def height(self):
                return component_class.height.fget(self) + NARROW_INC

        return new_class

    return actual_decorator


def boxed(component_class):
    class new_class(component_class):
        def render(self, x, y, *args, **kwargs):
            return [
                RectangleModel(
                    x=x,
                    y=y,
                    width=component_class.width.fget(self) + 2 * PADDING,
                    height=component_class.height.fget(self) + 2 * PADDING,
                    stroke="#000000",
                )
            ] + component_class.render(self, x + PADDING, y + PADDING, *args, **kwargs)

        @property
        def width(self):
            return component_class.width.fget(self) + 2 * PADDING

        @property
        def height(self):
            return component_class.height.fget(self) + 2 * PADDING

    return new_class


class HorizontalLayout:
    def __init__(self, children=None, padding=PADDING):
        if children is None:
            self.children = []
        else:
            self.children = children
        self.padding = padding

    def render(self, x, y, align=Align.TOP):
        if align is not Align.TOP:
            height = self.height
        res = []
        for child in self.children:
            if align is Align.TOP:
                y_ = y
            elif align is Align.CENTER:
                y_ = y + (height - child.height) / 2
            elif align is Align.BOTTOM:
                y_ = y + (height - child.height)
            else:
                raise ValueError(f"Invalid align {align} for HorizontalLayout")
            res.extend(safe_render(child, x, y))
            x += child.width + self.padding
        return res

    @property
    def width(self):
        if not self.children:
            return 0
        return self.padding * (len(self.children) - 1) + sum(
            c.width for c in self.children
        )

    @property
    def height(self):
        if not self.children:
            return 0
        return max(c.height for c in self.children)


class VerticalLayout:
    def __init__(self, children=None, padding=PADDING):
        if children is None:
            self.children = []
        else:
            self.children = children
        self.padding = padding

    def render(self, x, y):
        res = []
        for child in self.children:
            res.extend(safe_render(child, x, y))
            y += child.height + self.padding
        return res

    @property
    def width(self):
        if not self.children:
            return 0
        return max(c.width for c in self.children)

    @property
    def height(self):
        if not self.children:
            return 0
        return self.padding * (len(self.children) - 1) + sum(
            c.height for c in self.children
        )


class MaybeEditableRow(HorizontalLayout):
    def __init__(
        self,
        device_id,
        schema_hash,
        key_path,
        label_width=7 * NARROW_INC,
        display_width=5 * NARROW_INC,
        edit_width=5 * NARROW_INC,
        height=NARROW_INC,
    ):
        super().__init__(padding=0)
        key_attr = schema_hash.getAttributes(key_path)
        label_text = (
            key_attr["displayedName"]
            if "displayedName" in key_attr
            else key_path.split(".")[-1]
        )
        self.children.extend(
            [
                LabelModel(
                    text=label_text,
                    width=label_width,
                    height=height,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.{key_path}"],
                    width=display_width,
                    height=height,
                ),
            ]
        )
        if key_attr["accessMode"] == RECONFIGURABLE:
            value_type = key_attr["valueType"]
            if value_type in _type_to_line_editable:
                line_editable_class, extra_args = _type_to_line_editable[value_type]
                self.children.append(
                    line_editable_class(
                        keys=[f"{device_id}.{key_path}"],
                        width=edit_width,
                        height=height,
                        **extra_args,
                    )
                )
            else:
                self.children.append(
                    value_edit=LabelModel(
                        text=f"Not yet supported: editing {key_path} of type {value_type}",
                        width=edit_width,
                        height=height,
                    )
                )


@titled("Parameters used for CalCat queries", width=10 * NARROW_INC)
@boxed
class ConstantParameterColumn(VerticalLayout):
    def __init__(self, device_id, schema_hash, prefix="constantParameters"):
        super().__init__(padding=0)
        if "." in prefix:
            extra_path_prefix = prefix[: prefix.rfind(".") + 1]
        else:
            extra_path_prefix = ""
        self.children.extend(
            [
                MaybeEditableRow(
                    device_id,
                    schema_hash,
                    f"{prefix}.{key}",
                )
                for key in schema_hash.get(prefix).getKeys()
            ]
        )
        self.children.append(
            DisplayCommandModel(
                keys=[f"{device_id}.{extra_path_prefix}loadMostRecentConstants"],
                width=10 * BASE_INC,
                height=BASE_INC,
            )
        )


class ConstantNode(HorizontalLayout):
    def __init__(self, device_id, constant_path):
        super().__init__(padding=0)
        self.children.extend(
            [
                LabelModel(
                    text=constant_path.split(".")[-1],
                    width=7 * NARROW_INC,
                    height=NARROW_INC,
                ),
                ColorBoolModel(
                    width=NARROW_INC,
                    height=NARROW_INC,
                    keys=[f"{device_id}.{constant_path}.found"],
                ),
            ]
        )


@titled("Found constants", width=6 * NARROW_INC)
@boxed
class FoundConstantsColumn(VerticalLayout):
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        super().__init__(padding=0)
        self.children.extend(
            [
                ConstantNode(device_id, f"{prefix}.{constant_name}")
                for constant_name in schema_hash.get(prefix).getKeys()
            ]
        )


class Space:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def render(self, x, y):
        return []


class CorrectionStepNode(HorizontalLayout):
    def __init__(self, device_id, step_path):
        super().__init__(padding=0)
        self.children.extend(
            [
                LabelModel(
                    text=step_path.split(".")[-1], width=7 * BASE_INC, height=BASE_INC
                ),
                CheckBoxModel(
                    keys=[f"{device_id}.{step_path}.enable"],
                    width=BASE_INC,
                    height=BASE_INC,
                    klass="EditableCheckBox",
                ),
                CheckBoxModel(
                    keys=[f"{device_id}.{step_path}.preview"],
                    width=BASE_INC,
                    height=BASE_INC,
                    klass="EditableCheckBox",
                ),
            ]
        )


@titled("Correction steps", width=6 * NARROW_INC)
@boxed
class CorrectionStepsColumn(VerticalLayout):
    def __init__(self, device_id, schema_hash, prefix="corrections"):
        super().__init__(padding=0)
        self.children.append(
            HorizontalLayout(
                children=[
                    Space(width=7 * BASE_INC, height=BASE_INC),
                    LabelModel(
                        text="enable/preview", width=6 * BASE_INC, height=BASE_INC
                    ),
                ],
                padding=0,
            )
        )
        self.children.extend(
            [
                CorrectionStepNode(device_id, f"{prefix}.{step_path}")
                for step_path in schema_hash.get(prefix).getKeys()
            ]
        )


class ConstantLoadedAmpeln(HorizontalLayout):
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        super().__init__(padding=0)
        self.children.extend(
            [
                ColorBoolModel(
                    keys=[f"{device_id}.{prefix}.{key}.found"],
                    height=BASE_INC,
                    width=BASE_INC,
                )
                for key in schema_hash.get(prefix).getKeys()
            ]
        )


@titled("Manager status", width=6 * NARROW_INC)
@boxed
class ManagerDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        self.state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.restart_button = DisplayCommandModel(
            keys=[f"{device_id}.restartServers"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.instantiate_button = DisplayCommandModel(
            keys=[f"{device_id}.startInstantiate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.apply_button = DisplayCommandModel(
            keys=[f"{device_id}.applyManagedValues"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        self.children.extend(
            [
                self.name,
                HorizontalLayout(
                    children=[
                        self.state,
                        self.restart_button,
                    ],
                    padding=0,
                ),
                HorizontalLayout(
                    children=[self.instantiate_button, self.apply_button],
                    padding=0,
                ),
                self.status_log,
            ]
        )


@titled("Device status", width=6 * NARROW_INC)
@boxed
class CorrectionDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        self.state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.rate = EvaluatorModel(
            expression="f'{x:.02f}'",
            keys=[f"{device_id}.performance.rate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.processing_time = EvaluatorModel(
            expression="f'{x:.02f}'",
            keys=[f"{device_id}.performance.processingDuration"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.tid = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        self.children.extend(
            [
                self.name,
                HorizontalLayout(
                    children=[
                        self.state,
                        self.tid,
                    ],
                    padding=0,
                ),
                HorizontalLayout(
                    children=[
                        self.rate,
                        self.processing_time,
                    ],
                    padding=0,
                ),
                self.status_log,
            ]
        )


class CompactCorrectionDeviceOverview(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__(padding=0)
        self.children.extend(
            [
                DeviceSceneLinkModel(
                    text=device_id.split("/")[-1],
                    keys=[f"{device_id}.availableScenes"],
                    target="overview",
                    target_window=SceneTargetWindow.Dialog,
                    width=5 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayStateColorModel(
                    show_string=True,
                    keys=[f"{device_id}.state"],
                    width=6 * BASE_INC,
                    height=BASE_INC,
                ),
                EvaluatorModel(
                    expression="f'{x:.02f}'",
                    keys=[f"{device_id}.performance.rate"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.trainId"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                ConstantLoadedAmpeln(device_id, schema_hash),
            ]
        )


def correction_device_overview_scene(device_id, schema):
    if isinstance(schema, karathon.Schema):
        schema_hash = schema.getParameterHash()
    else:
        schema_hash = schema.hash

    content = HorizontalLayout(
        children=[
            CorrectionDeviceStatus(device_id),
            ConstantParameterColumn(device_id, schema_hash),
            VerticalLayout(
                children=[
                    FoundConstantsColumn(device_id, schema_hash),
                    CorrectionStepsColumn(device_id, schema_hash),
                ]
            ),
        ]
    )

    scene = SceneModel(
        children=content.render(PADDING, PADDING),
        width=content.width + 2 * PADDING,
        height=content.height + 2 * PADDING,
    )
    return write_scene(scene)


def manager_device_overview_scene(
    manager_device_id,
    manager_device_schema,
    correction_device_schema,
    correction_device_ids,
    domain_device_ids,
):
    mds_hash = schema_to_hash(manager_device_schema)
    cds_hash = schema_to_hash(correction_device_schema)

    content = VerticalLayout(
        children=[
            HorizontalLayout(
                children=[
                    ManagerDeviceStatus(manager_device_id),
                    VerticalLayout(
                        children=[
                            ConstantParameterColumn(
                                manager_device_id,
                                mds_hash,
                                prefix="managed.constantParameters",
                            ),
                            CorrectionStepsColumn(
                                manager_device_id,
                                mds_hash,
                                prefix="managed.corrections",
                            ),
                        ]
                    ),
                    titled("Other devices managed")(boxed(VerticalLayout))(
                        children=[
                            DeviceSceneLinkModel(
                                text=device_id.split("/")[-1],
                                keys=[f"{device_id}.availableScenes"],
                                width=14 * BASE_INC,
                                height=BASE_INC,
                            )
                            for device_id in sorted(
                                set(domain_device_ids)
                                - set(correction_device_ids)
                                - {manager_device_id}
                            )
                        ]
                    ),
                ],
            ),
            titled("Correction devices", width=8 * NARROW_INC)(boxed(VerticalLayout))(
                children=[
                    CompactCorrectionDeviceOverview(device_id, cds_hash)
                    for device_id in sorted(correction_device_ids)
                ],
                padding=0,
            ),
        ]
    )

    scene = SceneModel(
        width=content.width + 2 * PADDING,
        height=content.height + 2 * PADDING,
        children=content.render(PADDING, PADDING),
    )
    return write_scene(scene)


def schema_to_hash(schema):
    if isinstance(schema, karathon.Schema):
        return schema.getParameterHash()
    else:
        return schema.hash
